# greaseweazle/image/sf7.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

from ..image.img import IMG

class SF7(IMG):
    default_format = 'sega.sf7000'

# Local variables:
# python-indent: 4
# End:
