# greaseweazle/image/d81.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

from .img import IMG

class D81(IMG):
    default_format = 'commodore.1581'
    sides_swapped = True

# Local variables:
# python-indent: 4
# End:
